@@ -24,7 +24,9 @@ import socket
 import time
 
 import six
+import zipfile
 
+import pyghmi.constants as pygconst
 import pyghmi.exceptions as pygexc
 import pyghmi.ipmi.private.util as util
 import pyghmi.ipmi.oem.lenovo.config as config
@@ -34,6 +36,21 @@ import pyghmi.storage as storage
 from pyghmi.util.parse import parse_time
 import pyghmi.util.webclient as webclient
 
+class SensorReading(object):
+    def __init__(self, healthinfo, sensor=None, value=None, units=None,
+                 unavailable=False):
+        if sensor:
+            self.name = sensor['name']
+        else:
+            self.name = healthinfo['name']
+            self.health = healthinfo['health']
+            self.states = healthinfo['states']
+            self.state_ids = healthinfo.get('state_ids', None)
+        self.value = value
+        self.imprecision = None
+        self.units = units
+        self.unavailable = unavailable
+
 numregex = re.compile('([0-9]+)')
 funtypes = {
     0: 'RAID Controller',
@@ -85,7 +102,24 @@ def natural_sort(iterable):
         return sorted(iterable)
 
 
+def str_to_size(sizestr):
+    if 'GB' in sizestr:
+        sizestr = sizestr.replace('GB', '')
+        sizestr = int(float(sizestr) * 1000)
+    elif 'GiB' in sizestr:
+        sizestr = sizestr.replace('GiB', '')
+        sizestr = int(float(sizestr) * 1024)
+    elif 'TB' in sizestr:
+        sizestr = sizestr.replace('TB', '')
+        sizestr = int(float(sizestr) * 1000 * 1000)
+    elif 'TiB' in sizestr:
+        sizestr = sizestr.replace('TiB', '')
+        sizestr = int(float(sizestr) * 1024 * 1024)
+    return sizestr
+
+
 class OEMHandler(generic.OEMHandler):
+    usegenericsensors = False
     logouturl = '/api/providers/logout'
     bmcname = 'XCC'
     ADP_URL = '/api/dataset/imm_adapters?params=pci_GetAdapters'
@@ -116,6 +150,31 @@ class OEMHandler(generic.OEMHandler):
         self.fwc = None
         self.fwo = None
 
+    def get_screenshot(self, outfile):
+        self.wc.grab_json_response('/api/providers/rp_screenshot')
+        url = '/download/HostScreenShot.png'
+        fd = webclient.FileDownloader(self.wc, url, outfile)
+        fd.start()
+        fd.join()
+
+    def get_extended_bmc_configuration(self, fishclient, hideadvanced=True):
+        immsettings = self.get_system_configuration(fetchimm=True, hideadvanced=hideadvanced)
+        for setting in list(immsettings):
+            if not setting.startswith('IMM.'):
+                del immsettings[setting]
+        return immsettings
+
+    def get_ikvm_methods(self):
+        return ['url']
+
+    def get_ikvm_launchdata(self):
+        access = self._do_web_request('/redfish/v1/Managers/1/Oem/Lenovo/RemoteControl/Actions/LenovoRemoteControlService.GetRemoteConsoleToken', {})
+        if access.get('Token', None):
+            accessinfo = {
+                'url': '/#/login?{}&context=remote&mode=multi'.format(access['Token'])
+                }
+            return accessinfo
+
     def get_system_configuration(self, hideadvanced=True, fishclient=None,
                                  fetchimm=False):
         if not self.fwc:
@@ -236,6 +295,11 @@ class OEMHandler(generic.OEMHandler):
                                 '({2})'.format(
                                     newvalue, key,
                                     ','.join(self.fwo[key]['possible'])))
+                elif self.fwo[key]['validexpression']:
+                    if not re.match(self.fwo[key]['validexpression'], newvalue):
+                        raise pygexc.InvalidParameterValue(
+                            '"{0}" does not match expression "{1}"'.format(
+                                newvalue, self.fwo[key]['validexpression']))
                 newnewvalues.append(newvalue)
             if len(newnewvalues) == 1:
                 self.fwo[key]['new_value'] = newnewvalues[0]
@@ -246,8 +310,15 @@ class OEMHandler(generic.OEMHandler):
         if bay != -1:
             raise pygexc.UnsupportedFunctionality(
                 'This is not an enclosure manager')
-        rsp = self.wc.grab_json_response_with_status(
-            '/api/providers/virt_reseat', '{}')
+        wc = self.wc.dupe(timeout=5)
+        try:
+            rsp = wc.grab_json_response_with_status(
+                '/api/providers/virt_reseat', '{}')
+        except socket.timeout:
+            # Can't be certain, but most likely a timeout'
+            return
+        if rsp[1] == 500 and rsp[0] == 'Target Unavailable':
+            return
         if rsp[1] != 200 or rsp[0].get('return', 1) != 0:
             raise pygexc.UnsupportedFunctionality(
                 'This platform does not support AC reseat.')
@@ -477,7 +548,61 @@ class OEMHandler(generic.OEMHandler):
                     self.wc.grab_json_response(
                         '/api/function', {'USB_AddMapping': newmapping})
 
-    def get_description(self):
+    def get_health(self, fishclient, verbose=True):
+        rsp = self._do_web_request('/api/providers/imm_active_events')
+        summary = {'badreadings': [], 'health': pygconst.Health.Ok}
+        fallbackdata = []
+        hmap = {
+            'I': pygconst.Health.Ok,
+            'E': pygconst.Health.Critical,
+            'W': pygconst.Health.Warning,
+        }
+        infoevents = False
+        existingevts = set([])
+        for item in rsp.get('items', ()):
+            # while usually the ipmi interrogation shall explain things,
+            # just in case there is a gap, make sure at least the
+            # health field is accurately updated
+            itemseverity = hmap.get(item.get('severity', 'E'),
+                                    pygconst.Health.Critical)
+            if itemseverity == pygconst.Health.Ok:
+                infoevents = True
+                continue
+            if (summary['health'] < itemseverity):
+                summary['health'] = itemseverity
+            if item['cmnid'] == 'FQXSPPW0104J':
+                # This event does not get modeled by the sensors
+                # add a made up sensor to explain
+                fallbackdata.append(
+                    SensorReading({'name': item['source'],
+                                       'states': ['Not Redundant'],
+                                       'state_ids': [3],
+                                       'health': pygconst.Health.Warning,
+                                       'type': 'Power'}, ''))
+            elif item['cmnid'] == 'FQXSFMA0041K':
+                fallbackdata.append(
+                    SensorReading({
+                        'name': 'Optane DCPDIMM',
+                        'health': pygconst.Health.Warning,
+                        'type': 'Memory',
+                        'states': [item['message']]},
+                        '')
+                )
+            else:
+                currevt = '{}:{}'.format(item['source'], item['message'])
+                if currevt in existingevts:
+                    continue
+                existingevts.add(currevt)
+                fallbackdata.append(SensorReading({
+                    'name': item['source'],
+                    'states': [item['message']],
+                    'health': itemseverity,
+                    'type': item['source'],
+                }, ''))
+        summary['badreadings'] = fallbackdata
+        return summary
+
+    def get_description(self, fishclient):
         description = self._do_web_request('/DeviceDescription.json')
         if description:
             description = description[0]
@@ -493,8 +618,11 @@ class OEMHandler(generic.OEMHandler):
         return {'height': u_height, 'slot': slot}
 
     def _get_agentless_firmware(self, components):
+        skipkeys = set([])
         adata = self.wc.grab_json_response(
             '/api/dataset/imm_adapters?params=pci_GetAdapters')
+        fdata = self.wc.grab_json_response(
+            '/api/function/adapter_update?params=pci_GetAdapterListAndFW')
         anames = set()
         for adata in adata.get('items', []):
             baseaname = adata['adapterName']
@@ -527,11 +655,35 @@ class OEMHandler(generic.OEMHandler):
                         except ValueError:
                             pass
                     yield '{0} {1}'.format(aname, fname), bdata
+            for fwi in fdata.get('items', []):
+                if fwi.get('key', -1) == adata.get('key', -2):
+                    skipkeys.add(fwi['key'])
+                    if fwi.get('fw_status', 0) == 2:
+                        bdata = {}
+                        if 'fw_pkg_version' in fwi and fwi['fw_pkg_version']:
+                            bdata['version'] = fwi['fw_pkg_version']
+                        elif 'fw_version_pend' in fwi:
+                            bdata['version'] = fwi['fw_version_pend']
+                        yield '{0} Pending Update'.format(aname), bdata
+        for fwi in fdata.get('items', []):
+            if fwi.get('key', -1) > 0 and fwi['key'] not in skipkeys:
+                bdata = {}
+                bdata['version'] = fwi['fw_version']
+                yield fwi['adapterName'], bdata
+                if fwi.get('fw_status', 0) == 2:
+                    bdata = {}
+                    if 'fw_version_pend' in fwi:
+                        bdata['version'] = fwi['fw_version_pend']
+                    yield '{0} Pending Update'.format(fwi['adapterName']), bdata
 
     def _get_disk_firmware_single(self, diskent, prefix=''):
         bdata = {}
-        if not prefix and diskent.get('location', '').startswith('M.2'):
-            prefix = 'M.2-'
+        if not prefix:
+            location = diskent.get('location', '')
+            if location.startswith('M.2'):
+                prefix = 'M.2-'
+            elif location.startswith('7MM'):
+                prefix = '7MM-'
         diskname = 'Disk {1}{0}'.format(diskent['slotNo'], prefix)
         bdata['model'] = diskent[
             'productName'].rstrip()
@@ -547,7 +699,7 @@ class OEMHandler(generic.OEMHandler):
             for diskent in adp.get('aimDisks', ()):
                 yield self._get_disk_firmware_single(diskent)
 
-    def get_firmware_inventory(self, components):
+    def get_firmware_inventory(self, components, fishclient):
         sysinf = self.wc.grab_json_response('/api/dataset/sys_info')
         for item in sysinf.get('items', {}):
             for firm in item.get('firmware', []):
@@ -568,6 +720,8 @@ class OEMHandler(generic.OEMHandler):
                     yield ('LXPM Windows Driver Bundle', firminfo)
                 elif firm['type'] == 9:
                     yield ('LXPM Linux Driver Bundle', firminfo)
+                elif firm['type'] == 10:
+                    yield ('LXUM', firminfo)
         for adpinfo in self._get_agentless_firmware(components):
             yield adpinfo
         for adpinfo in self._get_disk_firmware(components):
@@ -606,12 +760,8 @@ class OEMHandler(generic.OEMHandler):
                             spares.append(diskinfo)
                         else:
                             disks.append(diskinfo)
-                    totalsize = pool['totalCapacityStr'].replace(
-                        'GB', '').replace('GiB', '')
-                    totalsize = int(float(totalsize) * 1024)
-                    freesize = pool['freeCapacityStr'].replace(
-                        'GB', '').replace('GiB', '')
-                    freesize = int(float(freesize) * 1024)
+                    totalsize = str_to_size(pool['totalCapacityStr'])
+                    freesize = str_to_size(pool['freeCapacityStr'])
                     pools.append(storage.Array(
                         disks=disks, raid=pool['rdlvlstr'], volumes=volumes,
                         id=(cid, pool['id']), hotspares=spares,
@@ -702,6 +852,19 @@ class OEMHandler(generic.OEMHandler):
             themap[raidname] = mapdata
         return themap
 
+    def check_storage_configuration(self, cfgspec=None):
+        rsp = self.wc.grab_json_response(
+            '/api/function/raid_conf?params=raidlink_GetStatus')
+        if rsp['items'][0]['status'] not in (2, 3):
+            raise pygexc.TemporaryError('Storage configuration unavailable in '
+                                        'current state (try boot to setup or '
+                                        'an OS)')
+        if not cfgspec:
+            return True
+        for pool in cfgspec.arrays:
+            self._parse_storage_cfgspec(pool)
+        return True
+
     def _wait_storage_async(self):
         rsp = {'items': [{'status': 0}]}
         while rsp['items'][0]['status'] == 0:
@@ -775,12 +938,23 @@ class OEMHandler(generic.OEMHandler):
     def _create_array(self, pool):
         params = self._parse_array_spec(pool)
         cid = params['controller'].split(',')[0]
+        cslotno = params['controller'].split(',')[1]
         url = '/api/function/raid_conf?params=raidlink_GetDefaultVolProp'
         args = (url, cid, 0, params['drives'])
         props = self.wc.grab_json_response(','.join([str(x) for x in args]))
-        if not props:  # newer firmwarerequires raidlevel too
+        usesctrlslot = False
+        if not props:  # newer firmware requires raidlevel too
             args = (url, cid, params['raidlevel'], 0, params['drives'])
             props = self.wc.grab_json_response(','.join([str(x) for x in args]))
+        elif 'return' in props and props['return'] == 22:
+            # Jan 2023 XCC FW - without controller slot number
+            args = (url, cid, params['raidlevel'], 0, params['drives'])
+            props = self.wc.grab_json_response(','.join([str(x) for x in args]))
+            if 'return' in props and props['return'] == 22:
+                usesctrlslot = True
+                # Jan 2023 XCC FW - with controller slot number
+                args = (url, cid, params['raidlevel'], 0, params['drives'], cslotno)
+                props = self.wc.grab_json_response(','.join([str(x) for x in args]))
         props = props['items'][0]
         volumes = pool.volumes
         remainingcap = params['capacity']
@@ -852,14 +1026,14 @@ class OEMHandler(generic.OEMHandler):
         parms = {'raidlink_AddNewVolWithNaAsync': arglist}
         rsp = self.wc.grab_json_response(url, parms)
         if rsp['return'] == 14:  # newer firmware
-            if 'supported_cpwb' in props: # Whitley
+            if 'supported_cpwb' in props and not usesctrlslot: # no ctrl_type
                 arglist = '{0},{1},{2},{3},{4},{5},{6},'.format(
                     cnum, params['raidlevel'], params['spans'],
                     params['perspan'], 0, params['drives'], params['hotspares'])
                 arglist += ''.join(vols)
                 parms = {'raidlink_AddNewVolWithNaAsync': arglist}
                 rsp = self.wc.grab_json_response(url, parms)
-            else: # Purley
+            else: # with ctrl_type
                 if cid[2] == 2:
                     cnum = cid[1]
                 arglist = '{0},{1},{2},{3},{4},{5},'.format(
@@ -977,7 +1151,7 @@ class OEMHandler(generic.OEMHandler):
         res = wc.grab_json_response_with_status(url, body, method=method)
         return res
 
-    def list_media(self, fishclient):
+    def list_media(self, fishclient, cache=True):
         rt = self.wc.grab_json_response('/api/providers/rp_vm_remote_getdisk')
         if 'items' in rt:
             for mt in rt['items']:
@@ -1074,6 +1248,29 @@ class OEMHandler(generic.OEMHandler):
         if usd['HttpPushUriTargetsBusy']:
             raise pygexc.TemporaryError('Cannot run multtiple updates to '
                                         'same target concurrently')
+        z = None
+        wrappedfilename = None
+        uxzcount = 0
+        needseek = False
+        if data and hasattr(data, 'read'):
+            if zipfile.is_zipfile(data):
+                needseek = True
+                z = zipfile.ZipFile(data)
+            else:
+                data.seek(0)
+        elif data is None and zipfile.is_zipfile(filename):
+            z = zipfile.ZipFile(filename)
+        if z:
+            for tmpname in z.namelist():
+                if tmpname.startswith('payloads/'):
+                    uxzcount += 1
+                    if tmpname.endswith('.uxz'):
+                        wrappedfilename = tmpname
+        if uxzcount == 1 and wrappedfilename:
+            filename = os.path.basename(wrappedfilename)
+            data = z.open(wrappedfilename)
+        elif needseek:
+            data.seek(0)
         upurl = usd['HttpPushUri']
         self._do_web_request(
             '/redfish/v1/UpdateService',
@@ -1109,25 +1306,46 @@ class OEMHandler(generic.OEMHandler):
                     raise Exception(uploadthread.rsp)
                 raise Exception(errmsg)
             rsp = json.loads(uploadthread.rsp)
-            monitorurl = rsp['TaskMonitor']
+            monitorurl = rsp['@odata.id']
             complete = False
-            while not complete:
-                pgress = self._do_web_request(monitorurl, cache=False)
+            phase = "apply"
+            statetype = 'TaskState'
+            # sometimes we get an empty pgress when transitioning from the apply phase to
+            # the validating phase; add a retry here so we don't exit the loop in this case
+            retry = 3
+            while not complete and retry > 0:
+                try:
+                    pgress = self._do_web_request(monitorurl, cache=False)
+                except socket.timeout:
+                    pgress = None
                 if not pgress:
-                    break
+                    retry -= 1
+                    time.sleep(3)
+                    continue
+                retry = 3 # reset retry counter
                 for msg in pgress.get('Messages', []):
                     if 'Verify failed' in msg.get('Message', ''):
                         raise Exception(msg['Message'])
-                state = pgress['TaskState']
+                state = pgress[statetype]
                 if state in ('Cancelled', 'Exception', 'Interrupted',
                              'Suspended'):
                     raise Exception(
                         json.dumps(json.dumps(pgress['Messages'])))
                 pct = float(pgress['PercentComplete'])
                 complete = state == 'Completed'
-                progress({'phase': 'apply', 'progress': pct})
-                if not complete:
+                progress({'phase': phase, 'progress': pct})
+                if complete:
+                    msgs = pgress.get('Messages', [])
+                    if msgs and 'OperationTransitionedToJob' in msgs[0].get('MessageId', ''):
+                        monitorurl = pgress['Messages'][0]['MessageArgs'][0]
+                        phase = 'validating'
+                        statetype = 'JobState'
+                        complete = False
+                        time.sleep(3)
+                else:
                     time.sleep(3)
+            if not retry:
+                raise Exception('Falied to monitor update progress due to excessive timeouts')
             if bank == 'backup':
                 return 'complete'
             return 'pending'
@@ -1342,8 +1560,12 @@ class OEMHandler(generic.OEMHandler):
         return 'pending'
 
     def get_diagnostic_data(self, savefile, progress=None, autosuffix=False):
-        self.wc.grab_json_response('/api/providers/ffdc',
-                                   {'Generate_FFDC': 1})
+        rsp = self.wc.grab_json_response('/api/providers/ffdc',
+                                         {'Generate_FFDC': 1})
+        if rsp.get('return', 0) == 4:
+            rsp = self.wc.grab_json_response('/api/providers/ffdc',
+                                             {'Generate_FFDC': 1,
+                                              'thermal_log': 0})
         percent = 0
         while percent != 100:
             time.sleep(3)
@@ -1373,7 +1595,7 @@ class OEMHandler(generic.OEMHandler):
             progress({'phase': 'complete'})
         return savefile
 
-    def get_licenses(self):
+    def get_licenses(self, fishclient):
         licdata = self.wc.grab_json_response('/api/providers/imm_fod')
         for lic in licdata.get('items', [{}])[0].get('keys', []):
             if lic['status'] == 0:
@@ -1384,7 +1606,7 @@ class OEMHandler(generic.OEMHandler):
                     'state': 'Missing required license'
                 }
 
-    def delete_license(self, name):
+    def delete_license(self, name, fishclient):
         licdata = self.wc.grab_json_response('/api/providers/imm_fod')
         for lic in licdata.get('items', [{}])[0].get('keys', []):
             if lic.get('feature', None) == name:
@@ -1397,7 +1619,7 @@ class OEMHandler(generic.OEMHandler):
                 )
                 break
 
-    def save_licenses(self, directory):
+    def save_licenses(self, directory, fishclient):
         licdata = self.wc.grab_json_response('/api/providers/imm_fod')
         for lic in licdata.get('items', [{}])[0].get('keys', []):
             licid = ','.join((str(lic['type']), str(lic['id'])))
@@ -1414,7 +1636,7 @@ class OEMHandler(generic.OEMHandler):
                     self._refresh_token()
                 yield savefile
 
-    def apply_license(self, filename, progress=None, data=None):
+    def apply_license(self, filename, fishclient, progress=None, data=None):
         license_errors = {
             310: "License is for a different model of system",
             311: "License is for a different system serial number",
@@ -1438,9 +1660,9 @@ class OEMHandler(generic.OEMHandler):
             if rsp.get('return', 0) in license_errors:
                 raise pygexc.InvalidParameterValue(
                     license_errors[rsp['return']])
-        return self.get_licenses()
+        return self.get_licenses(fishclient)
 
-    def user_delete(self, uid):
+    def user_delete(self, uid, fishclient=None):
         userinfo = self.wc.grab_json_response('/api/dataset/imm_users')
         uidtonamemap = {}
         for user in userinfo.get('items', [{'users': []}])[0].get('users', []):
@@ -1466,6 +1688,11 @@ class OEMHandler(generic.OEMHandler):
         yield "System"
         for key in natural_sort(hwmap):
             yield key
+        for cpuinv in self._get_cpu_inventory():
+            yield cpuinv[0]
+        for meminv in self._get_mem_inventory():
+            yield meminv[0]
+
 
     def get_inventory_of_component(self, compname):
         if compname.lower() == 'system':
@@ -1473,7 +1700,7 @@ class OEMHandler(generic.OEMHandler):
                 'UUID': self._varsysinfo.get('UUID', ''),
                 'Serial Number': self._varsysinfo.get('SerialNumber', ''),
                 'Manufacturer': self._varsysinfo.get('Manufacturer', ''),
-                'Product Name': self._varsysinfo.get('Model', ''),
+                'Product name': self._varsysinfo.get('Model', ''),
                 'Model': self._varsysinfo.get(
                     'SKU', self._varsysinfo.get('PartNumber', '')),
             }
@@ -1482,7 +1709,12 @@ class OEMHandler(generic.OEMHandler):
         try:
             return hwmap[compname]
         except KeyError:
-
+            for cpuinv in self._get_cpu_inventory():
+                if cpuinv[0] == compname:
+                    return cpuinv[1]
+            for meminv in self._get_mem_inventory():
+                if meminv[0] == compname:
+                    return meminv[1]
             return None
 
     def get_inventory(self, withids=False):
@@ -1490,7 +1722,7 @@ class OEMHandler(generic.OEMHandler):
             'UUID': self._varsysinfo.get('UUID', ''),
             'Serial Number': self._varsysinfo.get('SerialNumber', ''),
             'Manufacturer': self._varsysinfo.get('Manufacturer', ''),
-            'Product Name': self._varsysinfo.get('Model', ''),
+            'Product name': self._varsysinfo.get('Model', ''),
             'Model': self._varsysinfo.get(
                 'SKU', self._varsysinfo.get('PartNumber', '')),
         }
@@ -1558,7 +1790,7 @@ class OEMHandler(generic.OEMHandler):
                     if venid is not None:
                         bdata['PCI Vendor ID'] = '{0:04x}'.format(venid)
                     devid = fundata.get(self.ADP_DEVID, None)
-                    if devid is not None:
+                    if devid is not None and 'PCIE Device ID' not in bdata:
                         bdata['PCI Device ID'] = '{0:04x}'.format(devid)
                     venid = fundata.get(self.ADP_SUBVENID, None)
                     if venid is not None:
@@ -1658,8 +1890,12 @@ class OEMHandler(generic.OEMHandler):
 
     def get_disk_hardware(self, diskent, prefix=''):
         bdata = {}
-        if not prefix and diskent.get('location', '').startswith('M.2'):
-            prefix = 'M.2-'
+        if not prefix:
+            location = diskent.get('location', '')
+            if location.startswith('M.2'):
+                prefix = 'M.2-'
+            elif location.startswith('7MM'):
+                prefix = '7MM-'
         diskname = 'Disk {1}{0}'.format(diskent['slotNo'], prefix)
         bdata['Model'] = diskent['productName'].rstrip()
         bdata['Serial Number'] = diskent['serialNo'].rstrip()
